(function(root) {
  "use strict";

  var _previous_spritey = root.Spritey,
      build_animations,
      validate_animation_config,
      spr,
      helper = {}, // to test internal helper
      is_undefined      = helper.is_undefined     = function(val) { return typeof(val) === "undefined"; },
      is_function       = helper.is_function      = function(val) { return typeof(val) === "function"; },
      is_boolean        = helper.is_boolean       = function(val) { return typeof(val) === "boolean"; },
      is_string         = helper.is_string        = function(val) { return typeof(val) === "string"; },
      is_number         = helper.is_number        = function(val) { return typeof(val) === "number"; },
      is_null_or_undef  = helper.is_null_or_undef = function(val) { return (typeof(val) === "undefined") || (val === null); };

  validate_animation_config = function(cfg) {    
    if(!is_undefined(cfg.frames) && 
      (!is_undefined(cfg.from) || !is_undefined(cfg.to))) throw new spr.Error.InvalidAnimationConfig("Either use 'frames' or ['from' and 'to']!");

    if( (is_undefined(cfg.from) && !is_undefined(cfg.to)) ||
        (!is_undefined(cfg.from) && is_undefined(cfg.to))) throw new spr.Error.InvalidAnimationConfig("You have to set 'from' and 'to', not only one of them!");
  }

  build_animations = helper.build_animations = function(config, target) {
    config || (config = {});
    target || (target = {});

    for(var k in config) {
      if(config.hasOwnProperty(k)) {
        var cfg = config[k] || {},
            val = {};

        validate_animation_config(cfg);

        target[k]   = val;
        val.name    = k;
        val.frames  = cfg.frames || [];
        if(!is_undefined(cfg.from)) val.frames
        val.loop    = (is_undefined(cfg.loop) || (cfg.loop === false)) ? false : (is_boolean(cfg.loop) ? 0 : cfg.loop);
        val.after   = cfg.after || false;
      }
    }

    return target;
  };

  spr = root.Spritey = (function() {

    var Spritey = function(target, config) {
          config || (config = {});

          this.target         = target;
          this.cell_w         = (config.cells && config.cells.width) || 50;
          this.cell_h         = (config.cells && config.cells.height) || 50;
          this.cols           = (config.cells && config.cells.cols) || 1;
          this.rows           = (config.cells && config.cells.rows) || 1;
          this.setFrame       = config.setFrame || spr.setFrame; // frame-setting function (given or global)
          this.anis           = build_animations(config.animations || {});
          this.per_tick       = Math.floor((config.fps || 24) / 1000);
          this.interval       = undefined;
          this.current        = {
            ani_name: undefined,
            index_in_ani: undefined,
            frame: undefined
          }
        },
        proto = Spritey.prototype;

    var _ensure_play = function() {
      if(!this.running()) {
        this.interval = root.setInterval(this.tick, this.per_tick);
      }
    };

    var _find_ani = function(ani_name) {
      var ani = this.anis[ani_name];
      if(!ani) throw new spr.Error.AnimationNotFound("Animation " + ani_name + " not found!");

      return ani;
    };

    proto.tick = function() {
      console.log("tick!", this);
    };

    proto.update = function() {
      // call setFrame for current offset
    };

    proto.to = proto.play = function(ani_name) {
      if(!is_undefined(ani_name) && (ani_name !== this.current.ani_name)) {
        // set new ani (if not already playing this ani)

        var ani = _find_ani.call(this, ani_name);

        this.current.ani_name = ani_name;
        this.current.index_in_ani = 0;
        this.update();

        // todo: actually set init frame of current animation!
      }
      _ensure_play.call(this);
    };

    proto.running = function() {
      return !is_undefined(this.interval);
    };

    proto.stop = function() {
      if(!is_undefined(this.interval)) {
        root.clearInterval(this.interval);
        this.interval = undefined;
      }
    };

    return Spritey;
  })();

  spr.helper = helper;

  spr.setFrame = function(target, x_offset, y_offset) {
    target.style.backgroundPosition = "-" + x_offset + "px -" + y_offset; // todo: test
  };

  spr.noConflict = function() {
    root.Spritey = _previous_spritey;
    return spr;
  };


  // ERRORS

  (function() {
    var create_error = function(name, default_msg) {
      var err = function(message) {
        this.name = name;
        this.message = (message || default_msg || "");
      };

      err.prototype = new Error();
      err.prototype.constructor = err;
      err.prototype.name = name;

      return err;
    };

    spr.Error = {
      AnimationNotFound: create_error("AnimationNotFoundError", "The animation you tried to play does not exist"),
      InvalidAnimationConfig: create_error("InvalidAnimationConfigError", "The animation configuration is invalid")
    };
  })()

})(this);

if(false) { // trying out desired config format:
  ({
    cells: {
      width: 50,
      height: 50,
      cols: 8,
      rows: 2
    },
    fps: 20,
    init: "stand",
    animations: {
      stand: {
        from: 0,
        to: 10,
        loop: true // = 0
      },
      run: {
        from: 10,
        to: 19,
        loop: 6
      },
      sit: {
        frames: [0,1,2,2,2,1,2,3]
      },
      stand_up: {
        from: 20,
        to: 30,
        after: "stand"
      }
    }
  })

}